require "math"

local btk = {}

-------------------
-- Sane defaults --
-------------------

lf = love.filesystem
ls = love.sound
la = love.audio
lp = love.physics
lt = love.thread
li = love.image
lg = love.graphics
lc = love.timer -- love.chrono :-)
lm = love.mouse
lr = love.math -- love.rekenen :-)))

----------------
-- Game utils --
----------------

function btk.inQuintic(t)
	return math.pow(t, 5);
end

function btk.outCubic(progress) 
	return math.pow(progress, 3) + -3 * math.pow(progress, 2) + 3 * progress;
end

function btk.angleDiff(current, target)
    return math.atan2(math.sin(target - current), math.cos(target - current))
end

function btk.rotateDirection(current, target)
    if current < target then
        if math.abs(current - target) < math.pi then
            return 1
        else
            return -1
        end
    else
        if math.abs(current - target) < math.pi then
            return -1
        else
            return 1
        end
    end
end

function btk.debugDraw(body, shape)
    lg.setColor(0, 255, 0)
    lg.polygon("line", body:getWorldPoints(shape:getPoints()))
end

------------------
-- String utils --
------------------

function string.startsWith(input, start)
   return string.sub(input, 1, string.len(start)) == start
end

function string.endsWith(input, tail)
   return tail == '' or string.sub(input, -string.len(tail)) == tail
end

-----------------
-- Table utils --
-----------------

function table.slice(tbl, first, last, step)
    local sliced = {}

    for i = first or 1, last or #tbl, step or 1 do
        sliced[#sliced+1] = tbl[i]
    end

    return sliced
end

-------------------
-- Metafunctions --
-------------------

function btk.getType(o)
    local mt = getmetatable(o)

    if mt ~= nil then
        local oType = mt.type

        if oType ~= nil then
            return oType
        end
    end

    return type(o)
end

function btk.clone(o)
    -- TODO: What if it's not a table
    if type(o) == "table" then
        local new = {}

        setmetatable(new, getmetatable(o))
        for k, v in pairs(o) do
            valType = type(v)
            if valType == "table" then
                new[k] = btk.clone(v)
            else
                new[k] = v
            end
        end

        return new
    else
        return o
    end
end

-----------------
-- defaultdict --
-----------------

function btk.defaultdict(defaultVal)
    local o = {}

    setmetatable(o, {
        type = "defaultdict",
        __index = function()
            return defaultVal
        end
    })

    return o
end

-----------------
-- nestedArray --
-----------------

function btk.nestedArray(n, ...)
    if n == nil then return nil end

    local arr = {}
    for i = 1, n do
        arr[i] = btk.nestedArray(...)
    end
    
    return arr
end

----------
-- Vec --
----------

btk.Vec = {type = "Vec"}
setmetatable(btk.Vec, {
    __call = function(self, ...)
        local o = {}
        setmetatable(o, btk.Vec)

        o.n = select("#", ...)

        for i = 1, o.n do
            o[i] = select(i, ...)
        end

        return o
    end
})

function btk.Vec:__index(key)
    if key == "x" then
        return self[1]
    elseif key == "y" then
        return self[2]
    elseif key == "z" then
        return self[3]
    elseif key == "w" then
        return self[4]
    end

    return btk.Vec[key]
end

function btk.Vec:__newindex(k, v)
    if k == "x" then
        self[1] = v
    elseif k == "y" then
        self[2] = v
    elseif k == "z" then
        self[3] = v
    elseif k == "w" then
        self[4] = v
    else
        rawset(self, k, v)
    end
end

function btk.Vec:__tostring()
    local result = "[Vec: "

    if self.n > 0 then result = result .. self[1] end

    for _, v in ipairs(table.slice(self, 2)) do
        result = result .. " " .. v
    end

    return result .. "]"
end

function btk.Vec:mag()
    local result = 0

    for _, v in ipairs(self) do
        result = result + math.pow(v, 2)
    end

    return math.sqrt(result)
end

function btk.Vec:mag2()
    local result = 0

    for _, v in ipairs(self) do
        result = result + math.pow(v, 2)
    end

    return result
end

function btk.Vec.__eq(a, b)
    if a.n ~= b.n then return false end

    local result = true

    for i = 1, a.n do
        result = result and (a[i] == b[i])
        if not result then break end
    end

    return result
end

function btk.Vec.__lt(a, b)
    return a:mag2() < b:mag2()
end

function btk.Vec.__le(a, b)
    return a:mag2() <= b:mag2()
end

function btk.Vec.__add(a, b)
    local vec = btk.clone(a)

    for i = 1, vec.n do
        vec[i] = vec[i] + b[i]
    end

    return vec
end

function btk.Vec.__sub(a, b)
    local vec = btk.clone(a)

    for i = 1, vec.n do
        vec[i] = vec[i] - b[i]
    end

    return vec
end

function btk.Vec.__unm(a)
    local vec = btk.clone(a)

    for i = 1, vec.n do
        vec[i] = -vec[i]
    end

    return vec
end

function btk.Vec.__mul(a, b)
    local vec = btk.clone(a)
    
    if btk.getType(b) == "number" then
        for i = 1, vec.n do
            vec[i] = vec[i] * b
        end
    else
        for i = 1, vec.n do
            vec[i] = vec[i] * b[i]
        end
    end

    return vec
end

function btk.Vec:right()
    return btk.Vec(self.y, -self.x)
end

function btk.Vec:left()
    return self:right():right():right()
end

function btk.Vec:norm()
    local vec = btk.clone(self)
    local mag = self:mag()

    for k, v in ipairs(vec) do
        vec[k] = v / mag
    end

    return vec
end

----------
-- AABB --
----------

btk.AABB = {type = "AABB"}
setmetatable(btk.AABB, {
    __call = function(self, x, y, w, h)
        local o = {}
        setmetatable(o, btk.AABB)

        o.x = x
        o.y = y
        o.w = w
        o.h = h

        o.sides = {}

        return o
    end
})

btk.AABR = {}
setmetatable(btk.AABR, {
    __call = function(self, top, right, bottom, left)
        local o = {}
        setmetatable(o, btk.AABB)

        o.x = 0
        o.y = 0
        o.w = 0
        o.h = 0

        o.top = top
        o.right = right
        o.bottom = bottom
        o.left = left

        o.sides = {}

        return o
    end
})

function btk.AABB:__index(key)
    if key == "left" then
        return self.x
    elseif key == "top" then
        return self.y
    elseif key == "bottom" then
        return self.y + self.h
    elseif key == "right" then
        return self.x + self.w
    elseif key == "cx" then
        return self.x + self.w / 2
    elseif key == "cy" then
        return self.y + self.h / 2
    end

    return btk.AABB[key]
end

function btk.AABB:__newindex(k, v)
    if k == "left" then
        local diffX = v - self.x
        self.x = self.x + diffX
        self.w = self.w - diffX
    elseif k == "top" then
        local diffY = v - self.y
        self.y = self.y + diffY
        self.h = self.w - diffY
    elseif k == "right" then
        self.w = v - self.x
    elseif k == "bottom" then
        self.h = v - self.y
    elseif k == "cx" then
        self:setCenter(v, self.cy)
    elseif k == "cy" then
        self:setCenter(self.cx, v)
    else
        rawset(self, k, v)
    end
end

function btk.AABB:center()
    return btk.Vec(
        self.x + self.w / 2,
        self.y + self.h / 2
    )
end

function btk.AABB:grow(...)
    local dx, dy = nil, nil
    local n = select("#", ...)
    local arg1, arg2 = select(1, ...), select(2, ...)

    if n == 1 then
        local argType = btk.getType(arg1)

        if argType == "number" then
            dx = arg1
            dy = arg1
        elseif argType == "Vec" then
            dx = arg1.x
            dy = arg1.y
        else
            assert(false, "Unexpected type " .. argType)
        end
    elseif n == 2 then
        dx = arg1
        dy = arg2
    end

    self.x = self.x - dx
    self.y = self.y - dy
    self.w = self.w + 2 * dx
    self.h = self.h + 2 * dy

    return self
end

function btk.AABB:draw(style)
    love.graphics.rectangle(style or "line", self.x, self.y, self.w, self.h);
end

function btk.AABB:halfWidths()
    return btk.Vec(self.w / 2, self.h / 2)
end

function btk.AABB:containsPoint(...)
    local x = nil
    local y = nil

    if select("#", ...) == 1 then
        local point = select(1, ...) 
        assert(btk.getType(point) == "Vec", "Vec expected")
        
        x = point.x
        y = point.y
    elseif select("#", ...) == 2 then
        x = select(1, ...)
        y = select(2, ...)
    end

    if x > self.left and
            y > self.top and
            x < self.right and
            y < self.bottom then
        return {
            left = x - self.left,
            right = self.right - x,
            top = y - self.top,
            bottom = self.bottom - y,
        }
    end

    return nil
end

function btk.AABB:setCenter(...)
    local n = select("#", ...)
    local arg1 = select(1, ...)
    local arg2 = select(2, ...)

    if n == 1 then
        assert(getType(arg1) == "Vec", "Unexpected type " .. getType(arg1))
        self.x = arg1.x - self.w / 2
        self.y = arg1.y - self.h / 2
    elseif n == 2 then
        self.x = arg1 - self.w / 2
        self.y = arg2 - self.h / 2
    end

    return self
end

function btk.AABB:__tostring()
    return "[AABB: " ..
        self.x .. " "
        .. self.y .. " "
        .. self.w .. " "
        .. self.h .. "]"
end

function btk.AABB:move(...)
    local n = select("#", ...)
    local arg1 = select(1, ...)
    local arg2 = select(2, ...)

    local x, y = 0, 0

    if n == 1 then
        vec = arg1
        assert(btk.getType(vec) == "Vec", "Unexpected type " .. btk.getType(vec))
        x = vec.x
        y = vec.y
    elseif n == 2 then
        x = arg1
        y = arg2
    end

    self.x = self.x + x
    self.y = self.y + y

    return self
end

function btk.AABB.collide(a, b)
    local sides = {}
    if next(a.sides) == nil then -- If sides is empty
        sides = {
            top = true,
            right = true,
            bottom = true,
            left = true
        }
    else
        sides = a.sides
    end
    
    local collisionBox = btk.clone(a):grow(b:halfWidths())
    local ds = collisionBox:containsPoint(b:center())
    local dVec = nil
    
    if ds then
        local other = btk.Vec(0, -ds.top)
        if sides.top then
            dVec = other
        end

        other = btk.Vec(ds.right, 0)
        if sides.right and (not dVec or other < dVec) then
            dVec = other
        end

        other = btk.Vec(0, ds.bottom)
        if sides.bottom and (not dVec or other < dVec) then
            dVec = other
        end

        other = btk.Vec(-ds.left, 0)
        if sides.left and (not dVec or other < dVec) then
            dVec = other
        end
    end

    return dVec
end

function btk.AABB.sweepCollide(a, b, v)
    assert(false, "Not yet implemented")
end

return btk
